#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
import pickle
from inspect import isclass
from io import BytesIO
from typing import TYPE_CHECKING, Any

import msgpack
from bitarray import bitarray
from fsspec.implementations.local import LocalFileSystem
from upath import UPath

from .config import (
    BufferReaderType,
    config,
    decrement_gc_counter,
    increment_gc_counter,
)
from .index import normalise_index, to_index
from .unpacker import MsgspecUnpacker, Unpacker
from .writer import LazyWriter

if TYPE_CHECKING:
    from .config import BufferReader, FileSystem


def to_obj(v):
    """
    Ensure the given value is JSON serializable.
    """
    return v.to_obj() if isinstance(v, LazyItem) else v


async def async_to_obj(v):
    """
    Ensure the given value is JSON serializable.
    """
    if isinstance(v, LazyItem):
        return await asyncio.to_thread(v.to_obj)

    return v


async def async_get(v, key):
    """
    Get the value at the given key.
    """
    if isinstance(v, LazyItem):
        return await v.async_get(key)
    return v[key]


class LazyStats:
    def __init__(self):
        self._read_counter: int = 0
        self._call_counter: int = 0

    def __iadd__(self, other):
        self._read_counter += other
        self._call_counter += 1
        return self

    def __str__(self):
        return f"{self._call_counter} calls, {self._read_counter} bytes, {self.bytes_per_call()} bytes per call."

    def __call__(self, *args, **kwargs):
        return self._read_counter

    def bytes_per_call(self):
        return self._read_counter / self._call_counter

    def clear(self):
        self._read_counter = 0
        self._call_counter = 0


class LazyItem:
    def __init__(
        self,
        buffer: BufferReader,
        offset: int,
        *,
        counter: LazyStats | None = None,
        cached: bool = True,
        unpacker: Unpacker | None = None,
    ):
        self._buffer: BufferReader = buffer
        self._offset: int = offset  # start of original data
        self._counter: LazyStats | None = counter
        self._cached: bool = cached
        self._unpacker: Unpacker
        if isinstance(unpacker, Unpacker):
            self._unpacker = unpacker
        elif isclass(unpacker) and issubclass(unpacker, Unpacker):
            self._unpacker = unpacker()
        elif unpacker is None:
            self._unpacker = MsgspecUnpacker()
        else:
            raise TypeError("Need a valid unpacker.")

        self._accessed_items: int = 0

    def __len__(self):
        raise NotImplementedError

    def __eq__(self, other):
        return self.to_obj() == to_obj(other)

    def __str__(self):
        return self.to_obj().__str__()

    # noinspection SpellCheckingInspection
    def _readb(self, start: int, end: int):
        if self._buffer.closed:
            raise ValueError("File is closed.")

        size: int = end - start
        if self._counter:
            self._counter += size
        self._buffer.seek(start + self._offset)
        return self._buffer.read(size)

    def _unpack(self, data: bytes):
        return self._unpacker.decode(data)

    def _read(self, start: int, end: int):
        return self._unpack(self._readb(start, end))

    def _child(self, toc: dict | int):
        self._accessed_items += 1

        params: dict = {
            "counter": self._counter,
            "cached": self._cached,
            "unpacker": self._unpacker,
        }

        # {"t": {"name1": start_pos, "name2": start_pos}}
        # this is used in combined archives
        if isinstance(toc, int):
            self._buffer.seek(toc + self._offset)
            return LazyReader(self._buffer, **params)

        if (child_toc := toc.get("t", None)) is None:
            # {"p": [start_pos, end_pos]}
            # this is used in small objects
            if len(child_pos := toc["p"]) == 2 and all(
                isinstance(x, int) for x in child_pos
            ):
                if (
                    isinstance(data := self._read(*child_pos), bytes)
                    and b"multiarray" in data[:40]
                ):
                    return pickle.loads(data)
                return data

            # {"p": [[size1, start_pos, end_pos], [size2, start_pos, end_pos], [size3, start_pos, end_pos]]}
            # this is used in arrays of small objects
            return LazyList(toc, self._buffer, self._offset, **params)

        # {"t": [...], "p": [start_pos, end_pos]}
        # this is used in lazy lists
        if isinstance(child_toc, list):
            return LazyList(toc, self._buffer, self._offset, **params)

        # {"t": {...}, "p": [start_pos, end_pos]}
        # this is used in lazy dicts
        if isinstance(child_toc, dict):
            return LazyDict(toc, self._buffer, self._offset, **params)

        raise ValueError(f"Invalid: {toc}.")

    @property
    def _fast_loading(self):
        try:
            return (
                config.fast_loading
                and self._accessed_items < config.fast_loading_threshold * len(self)
            )
        except TypeError:
            return False

    def to_obj(self) -> Any:
        raise NotImplementedError

    def __getitem__(self, key):
        raise NotImplementedError

    async def async_get(self, key):
        return await asyncio.to_thread(self.__getitem__, key)


class LazyList(LazyItem):
    def __init__(
        self,
        toc: dict,
        buffer: BufferReader,
        offset: int,
        *,
        counter: LazyStats | None = None,
        cached: bool = True,
        unpacker: Unpacker | None = None,
    ):
        super().__init__(
            buffer, offset, counter=counter, cached=cached, unpacker=unpacker
        )
        self._toc: list = toc.get("t", None)  # noqa # if None, it's a list of small objects
        self._pos: list = toc.get("p", None)  # noqa # if None, it comes from a combined archive
        assert self._toc or self._pos, "Invalid TOC"
        self._index: int = 0
        self._cache: list = [None] * len(self)
        self._mask: bitarray = bitarray(len(self))
        self._mask.setall(0)  # ensure all bits are 0
        self._full_loaded: bool = False
        self._size_list: list = [0]
        if self._toc is None:
            total_size: int = 0
            for size, _, _ in self._pos:
                total_size += size
                self._size_list.append(total_size)

    def __repr__(self):
        return (
            f"LazyList[{len(self)}]"
            if config.simple_repr or not self._cached
            else self.to_obj().__repr__()
        )

    def _lookup_index(self, index: int) -> int:
        low: int = 0
        high: int = len(self._size_list) - 1

        while True:
            mid: int = (low + high) // 2

            if self._size_list[mid] <= index < self._size_list[mid + 1]:
                return mid

            if index < self._size_list[mid]:
                high = mid
            else:
                low = mid

    def _all(self, start: int, end: int) -> list:
        return list(msgpack.Unpacker(BytesIO(self._readb(start, end))))

    def __getitem__(self, index):
        index_range: list | range
        if isinstance(index, str):
            try:
                index_range = [int(index)]
            except ValueError as err:
                raise TypeError(
                    f"Invalid type: {type(index)} for index {index}."
                ) from err
        elif isinstance(index, slice):
            index_range = range(*index.indices(len(self)))
        elif isinstance(index, int):
            index_range = [index]
        else:
            raise TypeError(f"Invalid type: {type(index)} for index {index}.")

        if self._cached:
            for item in index_range:
                item = normalise_index(item, len(self))

                if self._mask[item] == 0:
                    if self._toc is not None:
                        self._mask[item] = 1
                        self._cache[item] = self._child(self._toc[item])
                    else:
                        lookup_index: int = self._lookup_index(item)
                        num_start, num_end = (
                            self._size_list[lookup_index],
                            self._size_list[lookup_index + 1],
                        )
                        self._mask[num_start:num_end] = 1
                        self._cache[num_start:num_end] = self._all(
                            *self._pos[lookup_index][1:]
                        )

            return self._cache[index]

        for item in index_range:
            item = normalise_index(item, len(self))

            if self._toc is not None:
                self._cache[item] = self._child(self._toc[item])
            else:
                lookup_index = self._lookup_index(item)
                num_start, num_end = (
                    self._size_list[lookup_index],
                    self._size_list[lookup_index + 1],
                )
                self._cache[num_start:num_end] = self._all(*self._pos[lookup_index][1:])

        result = self._cache[index]
        self._cache = [None] * len(self)
        return result

    def __iter__(self):
        self._index = 0
        return self

    def __next__(self):
        if self._index >= len(self):
            raise StopIteration

        item = self[self._index]
        self._index += 1
        return item

    def __len__(self):
        return (
            self._toc.__len__()
            if self._toc is not None
            else sum(x[0] for x in self._pos)
        )

    def to_obj(self):
        """
        Converts the data structure to a JSON serializable object.
        This method will read the entire data structure into memory.
        Data returned by this method can leave the `LazyReader` context.
        """

        def _read_all():
            for index in range(len(self)):
                self._cache[index] = to_obj(self[index])
            return self._cache

        if not self._cached:
            if self._toc is None:
                result: list = []
                for _, start, end in self._pos:
                    result.extend(self._all(start, end))
                return result

            return self._read(*self._pos) if self._pos else _read_all()

        if not self._full_loaded:
            self._full_loaded = True

            if not self._fast_loading:
                _read_all()
            elif self._toc is None:
                num_start, num_end = 0, 0
                for size, start, end in self._pos:
                    num_end += size
                    if self._mask[num_start] == 0:
                        self._cache[num_start:num_end] = self._all(start, end)
                    num_start = num_end
            elif self._pos:
                self._cache = self._read(*self._pos)
            else:
                _read_all()

            self._mask.setall(1)

        return self._cache


class LazyDict(LazyItem):
    def __init__(
        self,
        toc: dict,
        buffer: BufferReader,
        offset: int,
        *,
        counter: LazyStats | None = None,
        cached: bool = True,
        unpacker: Unpacker | None = None,
    ):
        super().__init__(
            buffer, offset, counter=counter, cached=cached, unpacker=unpacker
        )
        self._toc: dict = toc["t"]
        self._pos: list = toc.get("p", None)  # noqa # if empty, it comes from a combined archive
        self._cache: dict = {}
        self._full_loaded: bool = False

    def __repr__(self):
        return (
            f"LazyDict[{len(self)}]"
            if config.simple_repr or not self._cached
            else self.to_obj().__repr__()
        )

    def __getitem__(self, key):
        if not self._cached:
            return self._child(self._toc[key])

        if key not in self._cache:
            self._cache[key] = self._child(self._toc[key])

        return self._cache[key]

    def __contains__(self, item):
        return item in self._toc

    def __iter__(self):
        return self._toc.__iter__()

    def __len__(self):
        return self._toc.__len__()

    def get(self, key, default=None):
        """
        Mimics the `get` method for dictionaries.
        """
        return self[key] if key in self._toc else default

    def items(self):
        """
        Mimics the `items` method for dictionaries.
        """
        for k in self._toc:
            yield k, self[k]

    def keys(self):
        """
        Mimics the `keys` method for dictionaries.
        """
        return self._toc.keys()

    def values(self):
        """
        Mimics the `values` method for dictionaries.
        """
        for k in self._toc:
            yield self[k]

    def to_obj(self):
        """
        Converts the data structure to a JSON serializable object.
        This method will read the entire data structure into memory.
        Data returned by this method can leave the `LazyReader` context.
        """

        def _read_all():
            for k in self:
                self._cache[k] = to_obj(self[k])
            return self._cache

        if not self._cached:
            return self._read(*self._pos) if self._pos else _read_all()

        if not self._full_loaded:
            self._full_loaded = True
            if self._fast_loading and self._pos is not None:
                self._cache = self._read(*self._pos)
            else:
                _read_all()

        return self._cache


class LazyReader(LazyItem):
    def __init__(
        self,
        buffer_or_path: str | UPath | BufferReader,
        *,
        counter: LazyStats | None = None,
        cached: bool = True,
        unpacker: Unpacker | None = None,
        fs: FileSystem | None = None,
    ):
        """
        It is possible to use a customized unpacker.
        Please inherit the `Unpacker` class from the `unpacker.py`.
        There are already several unpackers available using different libraries.

        ```py
        class CustomUnpacker(Unpacker):
            def decode(self, data: bytes):
                # provide the decoding logic
                ...

        with LazyReader("file.msg", unpacker=CustomUnpacker()) as reader:
            # read the data
            ...
        ```

        There are a few different ways to read from a file.
        1. Provide `buffer_or_path` as a `str` with the path on the **local** filesystem.
        2. Provide `buffer_or_path` as a `str` with the path on the filesystem defined by `fs`.
        3. Provide `buffer_or_path` as a `BinaryIO` object that supports `.seek()` and `.read()` methods.
        4. Provide `buffer_or_path` as a `UPath` object that points to a file, it could be a local file or a remote one.

        It is recommended to simply pass in a `UPath` object that is generic to handle various underlying filesystems.

        :param buffer_or_path: the buffer or path to the file
        :param counter: the counter object for tracking the number of bytes read
        :param cached: whether to cache the data
        :param unpacker: the unpacker object for reading the data
        :param fs: `FileSystem` object for reading data from (if applicable)
        """
        self._buffer_or_path: str | UPath | BufferReader = buffer_or_path
        self._fs: FileSystem = fs or config.fs or LocalFileSystem()

        buffer: BufferReader
        if isinstance(self._buffer_or_path, str):
            buffer = self._fs.open(self._buffer_or_path, "rb", config.read_buffer_size)
        elif isinstance(self._buffer_or_path, UPath):
            buffer = self._buffer_or_path.open("rb", config.read_buffer_size)
        elif isinstance(self._buffer_or_path, BufferReaderType):
            buffer = self._buffer_or_path
        else:
            raise ValueError("Expecting a buffer or path.")

        sep_a, sep_b, sep_c = (
            LazyWriter.magic_len(),
            LazyWriter.magic_len() + 10,
            LazyWriter.magic_len() + 20,
        )

        # keep the buffer unchanged in case of failure
        original_pos: int = buffer.tell()
        header: bytes = buffer.read(sep_c)
        buffer.seek(original_pos)

        if not config.check_compatibility(header[:sep_a]):
            raise ValueError("Invalid file format.")

        super().__init__(
            buffer,
            original_pos + sep_c,
            counter=counter,
            cached=cached,
            unpacker=unpacker,
        )

        toc_start: int = self._unpack(header[sep_a:sep_b].lstrip(b"\0"))
        toc_size: int = self._unpack(header[sep_b:sep_c].lstrip(b"\0"))

        self._obj = self._child(self._read(toc_start, toc_start + toc_size))
        self._raw_data_range = (original_pos, sep_c + toc_start + toc_size)

    def __repr__(self):
        file_path: str = ""
        if isinstance(self._buffer_or_path, str):
            file_path = " (" + self._buffer_or_path + ")"

        return (
            f"LazyReader{file_path}"
            if config.simple_repr or not self._cached
            else self.to_obj().__repr__()
        )

    def __enter__(self):
        increment_gc_counter()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        decrement_gc_counter()

        if isinstance(self._buffer_or_path, (str, UPath)):
            self._buffer.close()

    def __getitem__(self, item):
        return self.read(item)

    def __len__(self):
        return len(self._obj)

    def __contains__(self, item):
        return item in self._obj

    def get(self, key, default=None):
        """
        Mimics the `get` method for dictionaries.
        """
        return self._obj.get(key, default)

    def keys(self):
        """
        Mimics the `keys` method for dictionaries.
        """
        return self._obj.keys()

    def values(self):
        """
        Mimics the `values` method for dictionaries.
        """
        return self._obj.values()

    def items(self):
        """
        Mimics the `items` method for dictionaries.
        """
        return self._obj.items()

    def read(self, path: str | list | slice | None = None):
        """
        Reads the data from the given path.

        This method navigates through the data structure based on the provided path.
        The path can be a string or a list. If it's a string, it's split into a list
        using '/' as the separator. Each element of the list is used to navigate
        through the data structure.

        If the path is None, it returns the root object.

        :param path: the path to the data to read
        :return: The data at the given path.
        """

        path_stack: list
        if path is None:
            path_stack = []
        elif isinstance(path, str):
            path_stack = path.split("/")
        elif isinstance(path, list):
            path_stack = path
        else:
            path_stack = [path]

        target = self._obj
        for key in (v for v in path_stack if v != ""):
            target = target[
                to_index(key, len(target))
                if isinstance(key, str) and isinstance(target, (list, LazyList))
                else key
            ]
        return target

    def visit(self, path: str = ""):
        """
        Reads the data from the given path.

        This method navigates through the data structure based on the provided path.
        The path can be a string of paths separated by '/'.

        If the path is None, it returns the root object.

        :param path: the path to the data to read
        :return: The data at the given path.
        """
        target = self._obj
        for key in (v for v in path.split("/") if v != ""):
            target = target[
                to_index(key, len(target))
                if isinstance(target, (list, LazyList))
                else key
            ]
        return target

    async def async_read(self, path: str | list | slice | None = None):
        """
        Reads the data from the given path.

        This method navigates through the data structure based on the provided path.
        The path can be a string or a list. If it's a string, it's split into a list
        using '/' as the separator. Each element of the list is used to navigate
        through the data structure.

        If the path is None, it returns the root object.

        :param path: the path to the data to read
        :return: The data at the given path.
        """

        path_stack: list
        if path is None:
            path_stack = []
        elif isinstance(path, str):
            path_stack = path.split("/")
        elif isinstance(path, list):
            path_stack = path
        else:
            path_stack = [path]

        target = self._obj
        for key in (v for v in path_stack if v != ""):
            target = await async_get(
                target,
                to_index(key, len(target))
                if isinstance(key, str) and isinstance(target, (list, LazyList))
                else key,
            )
        return target

    async def async_visit(self, path: str = ""):
        """
        Reads the data from the given path.

        This method navigates through the data structure based on the provided path.
        The path can be a string of paths separated by '/'.

        If the path is None, it returns the root object.

        :param path: the path to the data to read
        :return: The data at the given path.
        """
        target = self._obj
        for key in (v for v in path.split("/") if v != ""):
            target = await async_get(
                target,
                to_index(key, len(target))
                if isinstance(target, (list, LazyList))
                else key,
            )
        return target

    def to_obj(self):
        """
        Converts the data structure to a JSON serializable object.
        This method will read the entire data structure into memory.
        Data returned by this method can leave the `LazyReader` context.
        """
        return to_obj(self._obj)

    def raw_data(self):
        """
        Use to directly read the raw bytes out as a generator.
        For a single archive (generated by directly serialization instead of combining),
        this method is equivalent to open the file in binary mode and read everything out.

        ```py
            with open('target_file.msg', 'rb') as f, LazyReader('target_file.msg') as reader:
                assert f.read() == b''.join(reader.raw_data())
        ```
        """
        start, remaining = self._raw_data_range
        self._buffer.seek(start)

        while remaining > 0:
            if not (chunk := self._buffer.read(min(config.copy_chunk_size, remaining))):
                break

            yield chunk
            remaining -= len(chunk)



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from functools import lru_cache


@lru_cache(maxsize=2**14)
def _is_index(key: str) -> int | None:
    try:
        return int(key)
    except ValueError:
        return None


@lru_cache(maxsize=2**14)
def normalise_index(index: int, total_size: int) -> int:
    while index < -total_size:
        index += total_size
    while index >= total_size:
        index -= total_size
    return index


@lru_cache(maxsize=2**14)
def _normalise_bound(index: int, total_size: int) -> int:
    while index < 1 - total_size:
        index += total_size
    while index >= 1 + total_size:
        index -= total_size
    return index


@lru_cache(maxsize=2**14)
def _is_slice(key: str, total_size: int) -> tuple | None:
    if ":" not in key:
        return None

    parts: list = list(key.split(":"))

    if len(parts) == 2:
        start = 0 if parts[0] == "" else _is_index(parts[0])
        stop = total_size if parts[1] == "" else _is_index(parts[1])

        if start is not None and stop is not None:
            start = normalise_index(start, total_size)
            stop = _normalise_bound(stop, total_size)
            return start, stop, 1

        return None

    if len(parts) == 3:
        start = 0 if parts[0] == "" else _is_index(parts[0])
        step = 1 if parts[1] == "" else _is_index(parts[1])
        stop = total_size if parts[2] == "" else _is_index(parts[2])

        if start is not None and step is not None and stop is not None:
            start = normalise_index(start, total_size)
            stop = _normalise_bound(stop, total_size)
            return start, stop, step

        return None

    return None


@lru_cache(maxsize=2**14)
def to_index(key: str, total_size: int):
    if (int_key := _is_index(key)) is not None:
        return int_key

    if slicing := _is_slice(key, total_size):
        return slice(*slicing)

    return key



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING

from .config import config

if TYPE_CHECKING:
    from collections.abc import Callable
    from io import BytesIO
    from typing import BinaryIO

    from msgpack import Packer  # type: ignore

try:
    import numpy  # type: ignore

    ndarray = numpy.ndarray  # type: ignore
except ImportError:
    ndarray = list  # type: ignore


@dataclass(slots=True)
class Node:
    t: dict | list | None
    p: dict | list
    s: bool = False


class TOC:
    def __init__(
        self, *, packer: Packer, buffer: BytesIO | BinaryIO, transform: Callable = None
    ):
        self._buffer: BytesIO | BinaryIO = buffer
        self._packer: Packer = packer
        self._initial_pos = self._buffer.tell()
        self._in_numpy_array: bool = False

        def plain_forward(obj):
            return obj

        self._transform: Callable = transform or plain_forward

    @property
    def _pos(self) -> int:
        return self._buffer.tell() - self._initial_pos

    def _pack(self, obj) -> Node:
        def _pack_bin(_obj: bytes) -> None:
            self._buffer.write(_obj)

        def _pack_obj(_obj) -> None:
            self._buffer.write(self._packer.pack(_obj))

        def _generate(_start: int) -> Node:
            _end = self._pos
            return Node(None, [_start, _end], _end <= _start + config.trivial_size)

        if not isinstance(obj, (Mapping, list, set, tuple, ndarray)):
            start_pos = self._pos
            _pack_obj(obj)
            return _generate(start_pos)

        current_level_is_numpy_array: bool = False

        def _resume_flag(output):
            if current_level_is_numpy_array:
                self._in_numpy_array = False
            return output

        if isinstance(obj, tuple):
            obj = list(obj)
        elif isinstance(obj, set):
            obj = sorted(obj)
        elif ndarray is not list and isinstance(obj, ndarray):
            if config.numpy_encoder:
                start_pos = self._pos
                _pack_obj(obj.dumps())
                return _generate(start_pos)

            obj = obj.tolist()

            current_level_is_numpy_array = True
            self._in_numpy_array = True

        start_pos = self._pos

        obj_toc: dict | list
        all_small_obj: bool
        if isinstance(obj, Mapping):
            _pack_bin(self._packer.pack_map_header(len(obj)))
            obj_toc = {}
            for k, v in self._transform(obj.items()):  # type: ignore
                _pack_obj(k)
                obj_toc[k] = self._pack(v)
            all_small_obj = all(v.s for v in obj_toc.values())
        elif isinstance(obj, list):
            _pack_bin(self._packer.pack_array_header(len(obj)))

            if (
                self._in_numpy_array
                and len(obj) > 0
                and (
                    isinstance(obj[0], float)
                    or (config.numpy_fast_int_pack and isinstance(obj[0], int))
                )
            ):
                list_start: int = self._pos

                for v in obj:
                    _pack_obj(v)

                if self._pos < start_pos + config.small_obj_optimization_threshold:
                    return _resume_flag(_generate(start_pos))

                # assuming homogeneous list
                # compute the groups using a cheaper method
                total_items: int = len(obj)
                item_size: int = (self._pos - list_start) // total_items
                if item_size * total_items == self._pos - list_start:
                    group_size: int = min(
                        total_items,
                        config.small_obj_optimization_threshold // item_size + 1,
                    )
                    numpy_groups: list = []
                    current_pos: int = list_start
                    while total_items != 0:
                        current_block: int = min(group_size, total_items)
                        numpy_groups.append(
                            (
                                current_block,
                                current_pos,
                                current_pos + current_block * item_size,
                            )
                        )
                        current_pos += current_block * item_size
                        total_items -= current_block

                    assert current_pos == self._pos

                    return _resume_flag(Node(None, numpy_groups))

                self._buffer.seek(start_pos)

            obj_toc = [self._pack(v) for v in self._transform(obj)]  # type: ignore
            all_small_obj = all(v.s for v in obj_toc)
        else:
            raise ValueError(f"Expecting dict or list, got {obj.__class__}.")

        if self._pos < start_pos + config.small_obj_optimization_threshold:
            return _resume_flag(_generate(start_pos))

        if all_small_obj:
            if isinstance(obj, Mapping) or len(obj) == 0:
                return _resume_flag(_generate(start_pos))

            groups: list = []
            accu_list: list = []
            accu_size: int = 0
            for v in obj_toc:
                accu_list.append(v)
                accu_size += v.p[1] - v.p[0]
                if accu_size > config.small_obj_optimization_threshold:
                    groups.append(
                        (len(accu_list), accu_list[0].p[0], accu_list[-1].p[1])
                    )
                    accu_list = []
                    accu_size = 0

            if accu_list:
                groups.append((len(accu_list), accu_list[0].p[0], accu_list[-1].p[1]))

            return _resume_flag(
                Node(None, groups) if len(groups) > 1 else _generate(start_pos)
            )

        return _resume_flag(Node(obj_toc, [start_pos, self._pos]))

    def pack(self, obj) -> dict:
        def _factory(_obj) -> dict:
            return {k: v for k, v in _obj if v and k != "s"}

        return asdict(self._pack(obj), dict_factory=_factory)



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os.path
from io import BufferedReader, BytesIO
from tempfile import TemporaryFile
from typing import TYPE_CHECKING

from msgpack import Packer, packb, unpackb  # type: ignore
from upath import UPath

from .config import (
    config,
    decrement_gc_counter,
    increment_gc_counter,
    max_magic_len,
)
from .toc import TOC

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Literal

    from .config import BufferReader, BufferWriter, FileSystem


def _upsert(source: BufferReader, target: str, fs):
    if not fs:
        return

    with fs.open(target, "wb", block_size=config.write_buffer_size) as s3_file:
        # now transfer the local cache to s3
        source.seek(0)
        while chunk := source.read(config.write_buffer_size):
            s3_file.write(chunk)


class LazyWriter:
    magic: bytes = b"msglc-2024".rjust(max_magic_len, b"\0")

    @classmethod
    def magic_len(cls) -> int:
        return len(cls.magic)

    @classmethod
    def set_magic(cls, magic: bytes):
        cls.magic = magic.rjust(max_magic_len, b"\0")

    def __init__(
        self,
        buffer_or_path: str | UPath | BufferWriter,
        packer: Packer = None,
        *,
        fs: FileSystem | None = None,
    ):
        """
        It is possible to provide a custom packer object to be used for packing the object.
        However, this packer must be compatible with the `msgpack` packer.

        The `buffer_or_path` can be
        1. a plain `str` pointing to a file on local filesystem, or a file on target filesystem if `fs` is provided,
        2. a `UPath` object that points to a file on supported filesystem (either local or remote),
        3. a `IO` object that has `.tell()`, `.seek()`, `.write()` methods, this object must have random write access.

        Warning:
        Some of backend filesystems only supports sequential write rather than random write thus not all filesystems
        supported by `UPath` can be used.
        One must always check if the target filesystem supports random write access.
        If not, the most generic approach is to provide a plain string for `buffer_or_path` and explicitly assign a `fs` object.
        In this case, a local cache will be used to temporarily handle the serialization and the binary blob
        will be uploaded to the remote once everything is processed.

        :param buffer_or_path: target buffer or file path
        :param packer: packer object to be used for packing the object
        :param fs: `FileSystem` object to be used for storing
        """
        self._buffer_or_path: str | UPath | BufferWriter = buffer_or_path
        self._packer: Packer = packer or Packer()
        self._fs: FileSystem | None = fs or config.fs

        self._buffer: BufferWriter | TemporaryFile = None  # type: ignore
        self._toc_packer: TOC = None  # type: ignore
        self._header_start: int = 0
        self._file_start: int = 0
        self._no_more_writes: bool = False

    def __enter__(self):
        increment_gc_counter()

        if isinstance(self._buffer_or_path, str):
            if self._fs:
                # we need to seek to the beginning and overwrite the header
                # however, s3 does not allow seek in write mode
                # thus use a local temp file as cache
                self._buffer = TemporaryFile()
            else:
                self._buffer = open(
                    self._buffer_or_path, "wb", config.write_buffer_size
                )
        elif isinstance(self._buffer_or_path, UPath):
            self._buffer = self._buffer_or_path.open("wb")
            if not self._buffer.seekable():
                raise ValueError(
                    f"The underlying filesystem of the given UPath ({self._buffer_or_path}) does not support random write."
                )
        elif isinstance(self._buffer_or_path, (BytesIO, BufferedReader)):
            self._buffer = self._buffer_or_path
        else:
            raise ValueError("Expecting a buffer or path.")

        self._buffer.write(self.magic)
        self._header_start = self._buffer.tell()
        self._buffer.write(b"\0" * 20)
        self._file_start = self._buffer.tell()

        self._toc_packer = TOC(packer=self._packer, buffer=self._buffer)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        decrement_gc_counter()

        if isinstance(self._buffer_or_path, str):
            _upsert(self._buffer, self._buffer_or_path, self._fs)

        if isinstance(self._buffer_or_path, (str, UPath)):
            self._buffer.close()

    def write(self, obj) -> None:
        """
        This function is used to write the object to the file.

        Only one write is allowed. The function raises a `ValueError` if it is called more than once.

        :param obj: the object to be written to the file
        :raise ValueError: if the function is called more than once
        :return: None
        """
        if self._no_more_writes:
            raise ValueError("No more writes allowed.")

        self._no_more_writes = True

        toc: dict = self._toc_packer.pack(obj)
        toc_start: int = self._buffer.tell() - self._file_start
        packed_toc: bytes = self._packer.pack(toc)

        self._buffer.write(packed_toc)
        self._buffer.seek(self._header_start)
        self._buffer.write(self._packer.pack(toc_start).rjust(10, b"\0"))
        self._buffer.write(self._packer.pack(len(packed_toc)).rjust(10, b"\0"))


class LazyCombiner:
    def __init__(
        self,
        buffer_or_path: str | UPath | BufferWriter,
        *,
        mode: Literal["a", "w"] = "w",
        fs: FileSystem | None = None,
    ):
        """
        The mode resembles typical mode designations and implies the same meaning.
        If the mode is 'w', the file is overwritten.
        If the mode is 'a', the file is appended.

        :param buffer_or_path: target buffer or file path
        :param mode: mode of operation, 'w' for write and 'a' for append
        :param fs: `FileSystem` object to be used for storing
        """
        self._buffer_or_path: str | UPath | BufferWriter = buffer_or_path
        self._mode: str = mode
        self._fs: FileSystem | None = fs or config.fs

        self._buffer: BufferWriter | TemporaryFile = None  # type: ignore

        self._toc: dict | list = None  # type: ignore
        self._header_start: int = 0
        self._file_start: int = 0

    def __enter__(self):
        if isinstance(self._buffer_or_path, str):
            # we do not use a unified interface via `LocalFileSystem`
            # because for local storage we can skip the cache file
            if self._fs:
                self._buffer = TemporaryFile()
                if self._fs.exists(self._buffer_or_path):
                    with self._fs.open(self._buffer_or_path, "rb") as s3_file:
                        while chunk := s3_file.read(config.read_buffer_size):
                            self._buffer.write(chunk)
                    self._buffer.seek(0)
            else:
                mode: str = (
                    "wb"
                    if not os.path.exists(self._buffer_or_path) or self._mode == "w"
                    else "r+b"
                )
                self._buffer = open(  # type: ignore
                    self._buffer_or_path, mode, config.write_buffer_size
                )
        elif isinstance(self._buffer_or_path, UPath):
            self._buffer = self._buffer_or_path.open(
                "wb"
                if not self._buffer_or_path.exists() or self._mode == "w"
                else "r+b"
            )
            if not self._buffer.seekable():
                raise ValueError(
                    f"The underlying filesystem of the given UPath ({self._buffer_or_path}) does not support random write."
                )
        elif isinstance(self._buffer_or_path, (BytesIO, BufferedReader)):
            self._buffer = self._buffer_or_path
            if self._mode == "a":
                # need to read the header anyway
                self._buffer.seek(0)
        else:
            raise ValueError("Expecting a buffer or path.")

        if self._mode == "w":
            self._buffer.write(LazyWriter.magic)
            self._header_start = self._buffer.tell()
            self._buffer.write(b"\0" * 20)
            self._file_start = self._buffer.tell()
        else:
            sep_a, sep_b, sep_c = (
                LazyWriter.magic_len(),
                LazyWriter.magic_len() + 10,
                LazyWriter.magic_len() + 20,
            )

            ini_position: int = self._buffer.tell()
            header: bytes = self._buffer.read(sep_c)

            def _raise_invalid(msg: str):
                self._buffer.seek(ini_position)
                raise ValueError(msg)

            if not config.check_compatibility(header[:sep_a]):
                _raise_invalid(
                    "Invalid file format, cannot append to the current file."
                )

            toc_start: int = unpackb(header[sep_a:sep_b].lstrip(b"\0"))
            toc_size: int = unpackb(header[sep_b:sep_c].lstrip(b"\0"))

            self._buffer.seek(ini_position + sep_c + toc_start)
            self._toc = unpackb(self._buffer.read(toc_size)).get("t", None)

            if isinstance(self._toc, list):
                if any(not isinstance(i, int) for i in self._toc):
                    _raise_invalid("The given file is not a valid combined file.")
            elif isinstance(self._toc, dict):
                if any(not isinstance(i, int) for i in self._toc.values()):
                    _raise_invalid("The given file is not a valid combined file.")
            else:
                _raise_invalid("The given file is not a valid combined file.")

            self._header_start = ini_position + sep_a
            self._file_start = ini_position + sep_c
            self._buffer.seek(ini_position + sep_c + toc_start)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        toc_start: int = self._buffer.tell() - self._file_start
        packed_toc: bytes = packb({"t": self._toc})

        self._buffer.write(packed_toc)
        self._buffer.seek(self._header_start)
        self._buffer.write(packb(toc_start).rjust(10, b"\0"))
        self._buffer.write(packb(len(packed_toc)).rjust(10, b"\0"))

        if isinstance(self._buffer_or_path, str):
            _upsert(self._buffer, self._buffer_or_path, self._fs)

        if isinstance(self._buffer_or_path, (str, UPath)):
            self._buffer.close()

    def write(self, obj: Generator, name: str | None = None) -> None:
        """
        Write a number of objects to the file.

        :param obj: a generator of objects to be written to the file
        :param name: a name to be assigned to the object, only required when combining in dict mode
        """
        if self._toc is None:
            self._toc = [] if name is None else {}

        if name is None:
            if not isinstance(self._toc, list):
                raise ValueError("Need a name when combining in dict mode.")
        else:
            if not isinstance(self._toc, dict):
                raise ValueError("Cannot assign a name when combining in list mode.")
            if name in self._toc:
                raise ValueError(f"File {name} already exists.")

        start: int = self._buffer.tell() - self._file_start
        for chunk in obj:
            self._buffer.write(chunk)

        if name is None:
            self._toc.append(start)
        else:
            self._toc[name] = start



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import nullcontext
from io import IOBase
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, BinaryIO
from uuid import uuid4

from fsspec.implementations.local import LocalFileSystem
from upath import UPath

from .config import config
from .reader import LazyReader, to_obj
from .writer import LazyCombiner, LazyWriter

if TYPE_CHECKING:
    from io import BytesIO
    from typing import Literal

    from .config import FileSystem


def dump(file: str | UPath | BytesIO, obj, **kwargs):
    """
    This function is used to write the object to the file.

    :param file: a string representing the file path
    :param obj: the object to be written to the file
    :param kwargs: additional keyword arguments to be passed to the `LazyWriter`
    :return: None
    """
    with LazyWriter(file, **kwargs) as msglc_writer:
        msglc_writer.write(obj)


class FileInfo:
    """
    Wrap the file path or in memory buffer and name into a FileInfo object.
    The `name` is optional and is only used when the file is combined in the dictionary (key-value) mode.

    It is possible to optionally pass in an in-memory object (rather than serialized bytes stream) to be serialized.
    In this case, `obj` shall be assigned, and `path` must be `None`.

    The `fs` can be different for each `FileInfo` object, meaning it is possible to combine files from different sources.
    It is not affected by the global `fs` object stored in `config`.

    :param path: a string representing the file path or an in memory buffer
    :param name: key name of the content in the combined dict
    :param fs: `FileSystem` object to read the object from
    :param obj: the object to be written to the file, when `path` is None, `obj` will be serialized.
    """

    def __init__(
        self,
        path: str | UPath | BinaryIO | LazyReader | None,
        name: str | None = None,
        *,
        fs: FileSystem | None = None,
        obj=None,
    ):
        # `path` and `obj` can both be None, but not both assigned at the same time
        if obj is not None:
            assert path is None, "When `obj` is assigned, `path` must be None."
        self.path = path
        self.name = name
        self._fs: FileSystem = fs or LocalFileSystem()
        self._obj = obj

    def _exists(self):
        if isinstance(self.path, str):
            return self._fs.exists(self.path)
        if isinstance(self.path, UPath):
            return self.path.exists()
        return True

    def _open(self):
        if isinstance(self.path, str):
            return self._fs.open(self.path)
        if isinstance(self.path, UPath):
            return self.path.open("rb")
        if isinstance(self.path, IOBase):
            return nullcontext(self.path)
        raise RuntimeError

    def validate(self):
        if isinstance(self.path, (str, UPath)):
            if not self._exists():
                raise ValueError(f"File {self.path} does not exist.")
            with self._open() as _file:
                if not config.check_compatibility(_file.read(LazyWriter.magic_len())):
                    raise ValueError(f"Invalid file format: {self.path}.")
        elif isinstance(self.path, IOBase):
            with self._open() as _file:
                ini_pos = _file.tell()
                magic = _file.read(LazyWriter.magic_len())
                _file.seek(ini_pos)
                if not config.check_compatibility(magic):
                    raise ValueError("Invalid file format.")

    def chunking(self):
        if isinstance(self.path, LazyReader):
            yield from self.path.raw_data()
        elif self.path is None:
            with TemporaryDirectory() as _tmp_dir:
                dump(file_path := UPath(_tmp_dir) / uuid4().hex, to_obj(self._obj))
                with LazyReader(file_path) as _tmp_file:
                    yield from _tmp_file.raw_data()
        else:
            with self._open() as _file:
                while _data := _file.read(config.copy_chunk_size):
                    yield _data


def combine(
    archive: str | UPath | BytesIO,
    files: FileInfo | list[FileInfo],
    *,
    mode: Literal["a", "w"] = "w",
    validate: bool = True,
    fs: FileSystem | None = None,
):
    """
    This function is used to combine the multiple serialized files into a single archive.
    If `fs` is given, the combined archive will be uploaded to remote.

    The files to be combined must exist in local filesystem regardless of whether `fs` is given.
    In other words, only local files can be combined.

    :param archive: a string representing the file path of the archive
    :param files: a list of FileInfo objects
    :param mode: a string representing the combination mode, 'w' for write and 'a' for append
    :param validate: switch on to validate the files before combining
    :param fs: `FileSystem` object to be used for storing
    :return: None
    """
    if isinstance(files, FileInfo):
        files = [files]

    if 0 < sum(1 for file in files if file.name is not None) < len(files):
        raise ValueError("Files must either all have names or all not have names.")

    if len(all_names := {file.name for file in files}) != len(files) and (
        len(all_names) != 1 or all_names.pop() is not None
    ):
        raise ValueError("Files must have unique names.")

    if validate:
        for file in files:
            file.validate()

    with LazyCombiner(archive, mode=mode, fs=fs) as combiner:
        for file in files:
            combiner.write(file.chunking(), file.name)


def append(
    archive: str | UPath | BytesIO,
    files: FileInfo | list[FileInfo],
    *,
    validate: bool = True,
    fs: FileSystem | None = None,
):
    """
    This function is used to append the multiple serialized files to an existing single archive.
    If `fs` is given, the target will be downloaded first if it exists in the remote.
    The final archive will be uploaded to remote.

    The files to be appended must exist in local filesystem regardless of whether `fs` is given.

    :param archive: a string representing the file path of the archive
    :param files: a list of FileInfo objects
    :param validate: switch on to validate the files before combining
    :param fs: `FileSystem` object to be used for storing
    :return: None
    """
    combine(archive, files, mode="a", validate=validate, fs=fs)



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import gc
from dataclasses import dataclass
from io import IOBase
from threading import Lock
from typing import TYPE_CHECKING, BinaryIO, Union

from msglc.utility import MockIO

BufferWriterType = (BinaryIO, IOBase)
BufferReaderType = BufferWriterType + (MockIO,)

if TYPE_CHECKING:
    from collections.abc import Callable

    from fsspec import AbstractFileSystem as FileSystem

    BufferWriter = Union[BinaryIO, IOBase]
    BufferReader = Union[BufferWriter, MockIO]


@dataclass
class Config:
    small_obj_optimization_threshold: int = 2**13  # 8KB
    write_buffer_size: int = 2**23  # 8MB
    read_buffer_size: int = 2**16  # 64KB
    fast_loading: bool = True
    fast_loading_threshold: float = 0.3
    trivial_size: int = 20
    disable_gc: bool = True
    simple_repr: bool = True
    copy_chunk_size: int = 2**24  # 16MB
    numpy_encoder: bool = False
    numpy_fast_int_pack: bool = False
    fs: FileSystem | None = None
    compatibility_check: Callable[[bytes], bool] | None = None

    def check_compatibility(self, magic: bytes):
        if self.compatibility_check is not None:
            return self.compatibility_check(magic)

        from .writer import LazyWriter

        return magic == LazyWriter.magic


config = Config()

max_magic_len: int = 30


def configure(
    *,
    small_obj_optimization_threshold: int | None = None,
    write_buffer_size: int | None = None,
    read_buffer_size: int | None = None,
    fast_loading: bool | None = None,
    fast_loading_threshold: int | float | None = None,
    trivial_size: int | None = None,
    disable_gc: bool | None = None,
    simple_repr: bool | None = None,
    copy_chunk_size: int | None = None,
    numpy_encoder: bool | None = None,
    numpy_fast_int_pack: bool | None = None,
    magic: bytes | None = None,
    fs: FileSystem | None = None,
    compatibility_check: Callable[[bytes], bool] | None = None,
):
    """
    This function is used to configure the settings. It accepts any number of keyword arguments.
    The function updates the values of the configuration parameters if they are provided in the arguments.

    :param small_obj_optimization_threshold:
            The threshold (in bytes) for small object optimization.
            Objects smaller than this threshold are not indexed.
    :param write_buffer_size:
            The size (in bytes) for the write buffer.
    :param read_buffer_size:
            The size (in bytes) for the read buffer.
    :param fast_loading:
            Flag to enable or disable fast loading.
            If enabled, the container will be read in one go, instead of reading each child separately.
    :param fast_loading_threshold:
            The threshold (0 to 1) for fast loading.
            With the fast loading flag turned on, fast loading will be performed if the number of
            already read children over the total number of children is smaller than this threshold.
    :param trivial_size:
            The size (in bytes) considered trivial, around a dozen bytes.
            Objects smaller than this size are considered trivial.
            For a list of trivial objects, the container will be indexed in a blocked fashion.
    :param disable_gc:
            Flag to enable or disable garbage collection.
    :param simple_repr:
            Flag to enable or disable simple representation used in the __repr__ method.
            If turned on, __repr__ will not incur any disk I/O.
    :param copy_chunk_size:
            The size (in bytes) for the copy chunk.
    :param numpy_encoder:
            Flag to enable or disable the `numpy` support.
            If enabled, the `numpy` arrays will be encoded using the `dumps` method provided by `numpy`.
            The arrays are stored as binary data directly.
            If disabled, the `numpy` arrays will be converted to lists before encoding.
    :param numpy_fast_int_pack:
            If enabled, the integer numpy array will be packed assigning each element has identical size (4 or 8 bytes).
            This improves the performance of packing by avoiding the overhead of checking the size of each element.
            However, depending on the backend, for example, `messagepack` C implementation packs unsigned long long or long long.
            But its python implementation packs integer of various lengths (1, 2, 3, 5, 9 bytes).
    :param magic:
            Magic bytes (max length: 30) to set, used to identify the file format version.
    :param fs:
            The global `FileSystem` object that will be used by default so that there is no need to provide this for every function call.
            It is used to 1) read data by readers, 2) write output by writers/combiners.
            To specify where combiners read input files **from**, assign a specific `FileSystem` object to each `FileInfo`.
    :param compatibility_check:
            A function that returns a boolean based on the input magic bytes.
    """
    if (
        isinstance(small_obj_optimization_threshold, int)
        and small_obj_optimization_threshold > 0
    ):
        config.small_obj_optimization_threshold = small_obj_optimization_threshold
        if config.trivial_size > config.small_obj_optimization_threshold:
            config.trivial_size = config.small_obj_optimization_threshold

    if isinstance(write_buffer_size, int) and write_buffer_size > 0:
        config.write_buffer_size = write_buffer_size

    if isinstance(read_buffer_size, int) and read_buffer_size > 0:
        config.read_buffer_size = read_buffer_size

    if isinstance(fast_loading, bool):
        config.fast_loading = fast_loading

    if (
        isinstance(fast_loading_threshold, (int, float))
        and 0 <= fast_loading_threshold <= 1
    ):
        config.fast_loading_threshold = fast_loading_threshold

    if isinstance(trivial_size, int) and trivial_size > 0:
        config.trivial_size = trivial_size
        if config.trivial_size > config.small_obj_optimization_threshold:
            config.small_obj_optimization_threshold = config.trivial_size

    if isinstance(disable_gc, bool):
        config.disable_gc = disable_gc

    if isinstance(simple_repr, bool):
        config.simple_repr = simple_repr

    if isinstance(copy_chunk_size, int) and copy_chunk_size > 0:
        config.copy_chunk_size = copy_chunk_size

    if isinstance(numpy_encoder, bool):
        config.numpy_encoder = numpy_encoder

    if isinstance(numpy_fast_int_pack, bool):
        config.numpy_fast_int_pack = numpy_fast_int_pack

    config.fs = fs

    config.compatibility_check = compatibility_check

    if isinstance(magic, bytes) and 0 < len(magic) <= max_magic_len:
        from msglc import LazyWriter

        LazyWriter.set_magic(magic)


_gc_counter: int = 0
_gc_lock = Lock()


def increment_gc_counter():
    global _gc_counter

    if not config.disable_gc:
        return _gc_counter

    with _gc_lock:
        if _gc_counter == 0:
            gc.disable()
        _gc_counter += 1

        return _gc_counter


def decrement_gc_counter():
    global _gc_counter

    if not config.disable_gc:
        return _gc_counter

    with _gc_lock:
        _gc_counter -= 1
        if _gc_counter < 0:
            raise RuntimeError(
                "GC counter underflow, check the call to increment_gc_counter() and decrement_gc_counter()."
            )
        if _gc_counter == 0:
            gc.enable()

        return _gc_counter



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from importlib.util import find_spec

import msgpack


class Unpacker(ABC):
    @abstractmethod
    def decode(self, data):
        raise NotImplementedError


class MsgpackUnpacker(Unpacker):
    def __init__(self):
        self._unpacker = msgpack.Unpacker()

    def decode(self, data):
        self._unpacker.feed(data)
        return self._unpacker.unpack()


if find_spec("msgspec"):
    import msgspec

    class MsgspecUnpacker(Unpacker):
        def __init__(self):
            self._unpacker = msgspec.msgpack.Decoder()

        def decode(self, data):
            return self._unpacker.decode(data)
else:
    MsgspecUnpacker = MsgpackUnpacker


if find_spec("ormsgpack"):
    import ormsgpack

    class OrmsgpackUnpacker(Unpacker):
        def decode(self, data):
            return ormsgpack.unpackb(data)
else:
    OrmsgpackUnpacker = MsgpackUnpacker



#  Copyright (C) 2024-2026 Theodore Chang
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from time import sleep
from typing import TYPE_CHECKING

from upath import UPath

if TYPE_CHECKING:
    from io import BytesIO


class MockIO:
    def __init__(
        self,
        path: str | UPath | BytesIO,
        mode: str,
        seek_delay: float = 0,
        read_speed: int | list = 1 * 2**20,
        fs=None,
    ):
        self._path = path if isinstance(path, (str, UPath)) else None
        if isinstance(path, str):
            if fs:
                self._io = fs.open(path, mode)
            else:
                self._io = open(path, mode)  # noqa: SIM115
        elif isinstance(path, UPath):
            self._io = path.open(mode)
        else:
            self._io = path
        self._seek_delay: float = seek_delay
        self._read_speed: int | list = read_speed

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _actual_speed(self, size):
        if isinstance(self._read_speed, int):
            return self._read_speed

        for min_size, speed in self._read_speed:
            if size < min_size:
                return speed

        return self._read_speed[-1][1]

    def tell(self):
        return self._io.tell()

    def seek(self, offset: int):
        sleep(self._seek_delay)
        self._io.seek(offset)

    def read(self, size: int):
        sleep(size / self._actual_speed(size))
        return self._io.read(size)

    def close(self):
        if self._path is not None:
            self._io.close()

    @property
    def closed(self):
        return self._io.closed
